# -*- coding: utf-8 -*-



# pylint: disable=wildcard-import
from udo.repl.compilation import *  # noqa: F401
from udo.repl.disassembly import disassemble  # noqa: F401
from udo.repl.inspections import all_inspections  # noqa: F401
from udo.repl.repl_builtins import get_var_dict_from_ctx  # noqa: F401
from udo.repl.scope import *  # noqa: F401
